/*	Direct3DDevice OpenGL driver
 *
 * Copyright 2001 TransGaming Technologies Inc.
 */
#include "config.h"

#include <assert.h>
#include <stdlib.h>

#define USE_NEW_HAL /* temporary */

#include "debugtools.h"
#include "ddrawi.h"
#include "d3dhal.h"
#include "d3d_private.h"
#include "ddraw/gl.h"
#include "dsurface/main.h"
#include "dsurface/dib.h"
#include "d3ddevice/main.h"
#include "d3ddevice/gl.h"

DEFAULT_DEBUG_CHANNEL(ddraw);

#define DIB3DDATA(surf) (DIB_PRIV(surf)->dib.d3d_data)
#define DIBGLFMT(surf) (DIB3DDATA(surf)[0])
#define DIBGLTYPE(surf) (DIB3DDATA(surf)[1])
#define DIBGLLEVEL(surf) (DIB3DDATA(surf)[2])
#define DIBGLDEVICE(surf) (DIB3DDATA(surf)[3])
#define DIBGLIFMT(surf) (DIB3DDATA(surf)[4])

#define TEXTURE_ID(surf) ((surf)->t.gbl_more.dwDriverReserved)

/* stuff we use that are still not in SGI's glext.h */
/* token values taken from http://oss.sgi.com/projects/ogl-sample/registry/ */
#ifndef GL_ARB_texture_border_clamp
#define GL_CLAMP_TO_BORDER_ARB	0x812D
#endif
#ifndef GL_ARB_texture_env_combine
#define GL_SUBTRACT_ARB		0x84E7
#endif
#ifndef GL_ARB_texture_env_dot3
#define GL_DOT3_RGB_ARB		0x86AE
#define GL_DOT3_RGBA_ARB	0x86AF
#endif
#ifndef GL_EXT_texture_env_dot3 /* ext 220 */
#define GL_DOT3_RGB_EXT		0x8740
#define GL_DOT3_RGBA_EXT	0x8741
#endif
#ifndef GL_IBM_mirrored_repeat /* ext 224 */
#define GL_MIRRORED_REPEAT_IBM	0x8370
#endif

static inline float GL_RSF(DWORD dwState)
{
    return *(float*)&dwState;
}

static inline void GL_RS2Col(GLfloat col[4], DWORD dwState)
{
    D3DCOLOR rcol = (D3DCOLOR)dwState;
    col[0] = RGBA_GETRED(rcol)   / 255.0;
    col[1] = RGBA_GETGREEN(rcol) / 255.0;
    col[2] = RGBA_GETBLUE(rcol)  / 255.0;
    col[3] = RGBA_GETALPHA(rcol) / 255.0;
}

/* Right now, this only supports the formats we enumerate. */
static BOOL GL_PixelFormat_To_TextureFormat(GL_Direct3DDeviceImpl *priv,
					    const DDPIXELFORMAT *pfmt,
					    GLenum *fmt,
					    GLint *ifmt,
					    GLenum *type)
{
    static const DWORD unsupported = DDPF_BUMPDUDV | DDPF_BUMPLUMINANCE
	| DDPF_PALETTEINDEXED1
	| DDPF_PALETTEINDEXED2 | DDPF_PALETTEINDEXED4 | DDPF_PALETTEINDEXED8
	| DDPF_PALETTEINDEXEDTO8 | DDPF_RGBTOYUV | DDPF_STENCILBUFFER
	| DDPF_YUV | DDPF_ZBUFFER | DDPF_ZPIXELS;

    /* And ALPHAPREMULT might cause trouble. */

    TRACE("surface format %08lx, %ld\n", pfmt->dwFlags, pfmt->u1.dwRGBBitCount);
    TRACE("masks: %08lx, %08lx, %08lx, %08lx\n", pfmt->u2.dwRBitMask, pfmt->u3.dwGBitMask, pfmt->u4.dwBBitMask, pfmt->u5.dwRGBAlphaBitMask);

    if (pfmt->dwFlags & unsupported)
	return FALSE;

    if (pfmt->dwFlags & DDPF_FOURCC)
    {
	DD_GL_ext ext;
	TRACE("fourcc: %4.4s\n", (const char*)&pfmt->dwFourCC);
	if (!HAVE_GL_EXT(priv, ARB_texture_compression))
	    return FALSE;
	ext = 0;
	switch (pfmt->dwFourCC) {
	case 0x31545844: /* DXT1 */
	    if (pfmt->dwFlags & DDPF_ALPHAPIXELS)
		*fmt = GL_COMPRESSED_RGBA_S3TC_DXT1_EXT;
	    else
		*fmt = GL_COMPRESSED_RGB_S3TC_DXT1_EXT;
	    ext = EXT_texture_compression_s3tc;
	    break;
	case 0x33545844: /* DXT3 */
	    *fmt = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
	    ext = EXT_texture_compression_s3tc;
	    break;
	case 0x35545844: /* DXT5 */
	    *fmt = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
	    ext = EXT_texture_compression_s3tc;
	    break;
	}
	if (!ext) {
	    FIXME("unsupported texture format %4.4s\n", (const char*)&pfmt->dwFourCC);
	    return FALSE;
	}
	if (!HAVE_GL_EXT(priv, ext))
	    return FALSE;
	*ifmt = *fmt;
	*type = 0;
	return TRUE;
    }

    if (pfmt->dwFlags & DDPF_COMPRESSED)
	return FALSE;

    if (pfmt->dwFlags & DDPF_RGB)
    {
	if (pfmt->dwFlags & (DDPF_LUMINANCE|DDPF_ALPHA))
	    return FALSE;

	if (pfmt->dwFlags & DDPF_ALPHAPIXELS)
	{
	    *ifmt = GL_RGBA;
	    *fmt = GL_BGRA;

	    if (pfmt->u1.dwRGBBitCount == 32)
	    {
		*type = GL_UNSIGNED_BYTE;
		return TRUE;
	    }
	    else if (pfmt->u1.dwRGBBitCount == 16)
	    {
		if (pfmt->u5.dwRGBAlphaBitMask == 0x8000)
		    *type = GL_UNSIGNED_SHORT_1_5_5_5_REV;
		else
		    *type = GL_UNSIGNED_SHORT_4_4_4_4_REV;

		return TRUE;
	    }
	    else
		return FALSE;
	}
	else
	{
	    *ifmt = GL_RGB;

	    if (pfmt->u1.dwRGBBitCount == 32)
	    {
		*fmt = GL_BGRA;
		*type = GL_UNSIGNED_BYTE;
		return TRUE;
	    }
	    else if (pfmt->u1.dwRGBBitCount == 16)
	    {
		if (pfmt->u2.dwRBitMask & 0x8000)
		{
		    *fmt = GL_RGB;
		    *type = GL_UNSIGNED_SHORT_5_6_5;
		}
		else
		{
		    *fmt = GL_BGRA;
		    *type = GL_UNSIGNED_SHORT_1_5_5_5_REV;
		}

		return TRUE;
	    }
	    else
		return FALSE;
	}
    }

    if (pfmt->dwFlags & DDPF_LUMINANCE)
    {
	if (pfmt->dwFlags & DDPF_ALPHAPIXELS)
	{
	    *fmt = GL_LUMINANCE_ALPHA;
	    *ifmt = GL_LUMINANCE_ALPHA;
	    *type = GL_UNSIGNED_BYTE;
	}
	else
	{
	    *fmt = GL_LUMINANCE;
	    *ifmt = GL_LUMINANCE;
	    *type = GL_UNSIGNED_BYTE;
	}

	return TRUE;
    }

    return FALSE;
}

#ifdef HAVE_GLUBUILD2DMIPMAPLEVELS
static unsigned int GL_count_levels(unsigned int width, unsigned int height)
{
    unsigned int x = max(width, height);
    unsigned int levels = 0;

    while (x > 1)
    {
	x /= 2;
	levels++;
    }

    return levels;
}
#endif

static HRESULT GL_PrepTexture(IDirect3DDeviceImpl *This, IDirectDrawSurfaceImpl *texture)
{
    GL_D3D_PRIV_VAR(priv, This);
    static DDSCAPS2 mipmap_caps = { DDSCAPS_TEXTURE | DDSCAPS_MIPMAP };
    LPDDPIXELFORMAT pfmt = &(texture->surface_desc.u4.ddpfPixelFormat);
    GLint ifmt;
    GLenum fmt, type;
    LPDIRECTDRAWSURFACE7 mipmap;
    int level;

    if (!TEXTURE_ID(texture)) {
	HRESULT hr = This->late_allocate_surface(This, texture);
	if (FAILED(hr)) return hr;
    }

    if (!GL_PixelFormat_To_TextureFormat(priv, pfmt, &fmt, &ifmt, &type))
    {
	FIXME("unhandled format %lx, %ld\n", pfmt->dwFlags,
	      pfmt->u1.dwRGBBitCount);
	return DDERR_INVALIDPARAMS;
    }

    glPushAttrib(GL_TEXTURE_BIT); /* save the current texture bindings */

    glBindTexture(GL_TEXTURE_2D, (GLuint)TEXTURE_ID(texture));

    mipmap = ICOM_INTERFACE(texture, IDirectDrawSurface7);
    IDirectDrawSurface7_AddRef(mipmap);

    level = 0;

    while (1)
    {
	IDirectDrawSurface7 *next;
	IDirectDrawSurfaceImpl *mipmapI;

	mipmapI = ICOM_OBJECT(IDirectDrawSurfaceImpl, IDirectDrawSurface7,
			      mipmap);

	if (type)
	    glTexImage2D(GL_TEXTURE_2D, level, ifmt,
			 mipmapI->surface_desc.dwWidth,
			 mipmapI->surface_desc.dwHeight,
			 0, fmt, type,
			 mipmapI->surface_desc.lpSurface);
	else
	    GLD(priv)->glCompressedTexImage2DARB(GL_TEXTURE_2D, level, ifmt,
						 mipmapI->surface_desc.dwWidth,
						 mipmapI->surface_desc.dwHeight,
						 0, mipmapI->surface_desc.u1.dwLinearSize,
						 mipmapI->surface_desc.lpSurface);

	DIBGLFMT(mipmapI) = fmt;
	DIBGLIFMT(mipmapI) = ifmt;
	DIBGLTYPE(mipmapI) = type;
	DIBGLLEVEL(mipmapI) = level;
	DIBGLDEVICE(mipmapI) = (DWORD)This;

	level++;

	/* Get the next mipmap. */
	if (FAILED(IDirectDrawSurface7_GetAttachedSurface(mipmap, &mipmap_caps,
							  &next)))
	{
	    level--; /* level is the max (smallest) mipmap level */

	    /* Fill in the remaining mipmaps as a workaround for drivers that
	     * do not support GL_TEXTURE_MAX_LEVEL.
	     *
	     * XXX Should we do this for compressed textures too?
	     *
	     * Note that old Mesa libGLU may not have this function.
	     * Get a recent version that includes the SGI libGLU.
	     */
#ifdef HAVE_GLUBUILD2DMIPMAPLEVELS
	    if (type
		&& (mipmapI->surface_desc.dwWidth > 1
		    || mipmapI->surface_desc.dwHeight > 1))
	    {
		DWORD max_level = level
		    + GL_count_levels(mipmapI->surface_desc.dwWidth,
				      mipmapI->surface_desc.dwHeight);

		gluBuild2DMipmapLevels(GL_TEXTURE_2D, ifmt,
				       mipmapI->surface_desc.dwWidth,
				       mipmapI->surface_desc.dwHeight,
				       fmt, type, level, level+1, max_level,
				       mipmapI->surface_desc.lpSurface);
	    }
#endif

	    IDirectDrawSurface7_Release(mipmap);
	    break;
	}

	IDirectDrawSurface7_Release(mipmap);
	mipmap = next;
    }

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, level);

    glPopAttrib();

    TRACE("initialized GL texture %d\n", TEXTURE_ID(texture));

    return D3D_OK;
}

#ifdef HAVE_GLUBUILD2DMIPMAPLEVELS
static BOOL
D3D_GL_is_last_mipmap(IDirectDrawSurfaceImpl* texture)
{
    static DDSCAPS2 mipmap_caps = { DDSCAPS_TEXTURE | DDSCAPS_MIPMAP };
    LPDIRECTDRAWSURFACE7 tex, next;

    tex = ICOM_INTERFACE(texture, IDirectDrawSurface7);

    if (FAILED(IDirectDrawSurface7_GetAttachedSurface(tex, &mipmap_caps,
						      &next)))
	return TRUE;

    IDirectDrawSurface7_Release(next);
    return FALSE;
}
#endif

HRESULT D3D_GL_texblt(IDirect3DDeviceImpl* This, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2TEXBLT dat = (LPD3DHAL_DP2TEXBLT)data1;
    IDirectDrawSurfaceImpl *texture = OBJECT_FROM_FIELD(IDirectDrawSurfaceImpl, t, (void*)dat->dwDDDestSurface);
    IDirectDrawSurfaceImpl *srctex = OBJECT_FROM_FIELD(IDirectDrawSurfaceImpl, t, (void*)dat->dwDDSrcSurface);
    GLint oldtex;

    TRACE("(%p): src=%p, dst=%p\n", This, srctex, texture);

    /* PreLoad? */
    if (!texture) {
	if (!DIBGLDEVICE(srctex))
	    return GL_PrepTexture(This, srctex);
	return D3D_OK;
    }

    /* Texture previously been loaded into this context? */
    if (DIBGLDEVICE(texture) != (DWORD)This) return D3D_OK;

    assert(srctex == NULL);

    /* OK, perform requested update */
    glGetIntegerv(GL_TEXTURE_BINDING_2D, &oldtex);
    glBindTexture(GL_TEXTURE_2D, TEXTURE_ID(texture));
    if (DIBGLTYPE(texture))
    {
	glTexSubImage2D(GL_TEXTURE_2D, DIBGLLEVEL(texture),
			dat->rSrc.left, dat->rSrc.top,
			dat->rSrc.right - dat->rSrc.left, dat->rSrc.bottom - dat->rSrc.top,
			DIBGLFMT(texture), DIBGLTYPE(texture), texture->surface_desc.lpSurface);

	/* Despite GL_TEXTURE_MAX_LEVEL, we actually load all the mipmap
	 * levels to support buggy hardware.
	 * If the smallest mipmap level is being reloaded, then we need to
	 * regenerate the mipmap levels from it. */
#ifdef HAVE_GLUBUILD2DMIPMAPLEVELS
	if (D3D_GL_is_last_mipmap(texture)
	    && (texture->surface_desc.dwWidth > 1
		|| texture->surface_desc.dwHeight > 1))
	
	{
	    DWORD max_level = DIBGLLEVEL(texture);
		+ GL_count_levels(texture->surface_desc.dwWidth,
				  texture->surface_desc.dwHeight);

	    gluBuild2DMipmapLevels(GL_TEXTURE_2D, DIBGLIFMT(texture),
				   texture->surface_desc.dwWidth,
				   texture->surface_desc.dwHeight,
				   DIBGLFMT(texture), DIBGLTYPE(texture),
				   DIBGLLEVEL(texture), DIBGLLEVEL(texture)+1,
				   max_level, texture->surface_desc.lpSurface);
	}
#endif

    }
    else
	FIXME("update compressed texture\n");
    glBindTexture(GL_TEXTURE_2D, oldtex);

    TRACE("updated GL texture %d\n", TEXTURE_ID(texture));

    return D3D_OK;
}

static inline BOOL GL_IsStageEnabled(IDirect3DDeviceImpl* This, DWORD dwStage)
{
    unsigned u;
    for (u=0; u<=dwStage; u++)
	if (This->d3dp.texstagestate[u][D3DTSS_COLOROP] <= D3DTOP_DISABLE)
	    return FALSE;
    return TRUE;
}

static inline void GL_StageEnabled(IDirect3DDeviceImpl* This, DWORD dwStage)
{
    GL_D3D_PRIV_VAR(priv, This);
    unsigned u;
    if (This->d3dp.texstagestate[dwStage][D3DTSS_TEXTUREMAP])
	glEnable(GL_TEXTURE_2D);
    for (u=dwStage+1; u<priv->gl.lpCaps->wMaxTextureBlendStages &&
		      This->d3dp.texstagestate[u][D3DTSS_COLOROP]>D3DTOP_DISABLE; u++) {
	if (This->d3dp.texstagestate[u][D3DTSS_TEXTUREMAP]) {
	    GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + u);
	    glEnable(GL_TEXTURE_2D);
	}
    }
}

static inline void GL_StageDisabled(IDirect3DDeviceImpl* This, DWORD dwStage)
{
    GL_D3D_PRIV_VAR(priv, This);
    unsigned u;
    if (This->d3dp.texstagestate[dwStage][D3DTSS_TEXTUREMAP])
	glDisable(GL_TEXTURE_2D);
    for (u=dwStage+1; u<priv->gl.lpCaps->wMaxTextureBlendStages &&
		      This->d3dp.texstagestate[u][D3DTSS_COLOROP]>D3DTOP_DISABLE; u++) {
	if (This->d3dp.texstagestate[u][D3DTSS_TEXTUREMAP]) {
	    GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + u);
	    glDisable(GL_TEXTURE_2D);
	}
    }
}

typedef const struct {
    DD_GL_ext ext;
    DWORD dwFlags;
    GLint scale, cmode, amode, src0, src1, src2;
} GL_TEXOP;

#define E_GL 0
#define E_EA EXT_texture_env_add
#define E_EC EXT_texture_env_combine
#define E_AC ARB_texture_env_combine
#define E_E3 EXT_texture_env_dot3
#define E_A3 ARB_texture_env_dot3

/* If a SOURCE bit is set, that src value tells that a D3D texop argument should be
 * mapped and substituted to the corresponding GL source. If not, the src value is
 * considered a GL token, and is used directly (with a default operand), if non-zero.
 * For instance, for color ops, if dwFlags = GLTF_SOURCE0, src0 = 0, will mean that
 * D3DTSS_COLORARG1 will be mapped into GL_SOURCE0_RGB_EXT. Similarly, src0 = 1 will
 * map D3DTSS_COLORARG2 into GL_SOURCE0_RGB_EXT instead. */
#define GLTF_SOURCE0	0x0001
#define GLTF_SOURCE1	0x0002
#define GLTF_SOURCE2	0x0004
#define GLTF_SOURCE3	0x0008
#define GLTF_SOURCEA	GLTF_SOURCE0|GLTF_SOURCE1
#define GLTF_SOURCEB	GLTF_SOURCE1|GLTF_SOURCE2
/* Modifiers, not implemented or well-defined yet. */
#define GLTF_C		0x0010	/* Use a constant */
#define GLTF_SA		0x0020	/* Special alpha treatment */

/* If you implement anything, also try to update caps bits in ddraw/gl.c */

/* this table is indexed by the table below */
static GL_TEXOP GL_tex_op_data[] = {
/* supported by OpenGL 1.2 */
    /* SELECTARG1 */			{E_GL, GLTF_SOURCE0, 1, GL_REPLACE, 0, 0},
    /* SELECTARG2 */			{E_GL, GLTF_SOURCE0, 1, GL_REPLACE, 0, 1},
    /* MODULATE */			{E_GL, GLTF_SOURCEA, 1, GL_MODULATE, 0, 0, 1},
/* supported by ARB_texture_env_add */
    /* ADD */				{E_EA, GLTF_SOURCEA, 1, GL_ADD, 0, 0, 1},
/* supported by ARB_texture_env_combine */
    /* SUBTRACT */			{E_AC, GLTF_SOURCEA, 1, GL_SUBTRACT_ARB, 0, 0, 1},
/* supported by EXT_texture_env_combine */
    /* MODULATE2X */			{E_EC, GLTF_SOURCEA, 2, GL_MODULATE, 0, 0, 1},
    /* MODULATE4X */			{E_EC, GLTF_SOURCEA, 4, GL_MODULATE, 0, 0, 1},
    /* ADDSIGNED */			{E_EC, GLTF_SOURCEA, 1, GL_ADD_SIGNED_EXT, 0, 0, 1},
    /* ADDSIGNED2X */			{E_EC, GLTF_SOURCEA, 2, GL_ADD_SIGNED_EXT, 0, 0, 1},
    /* BLENDDIFFUSEALPHA */		{E_EC, GLTF_SOURCEA, 1, GL_INTERPOLATE_EXT, 0, 0, 1, GL_PRIMARY_COLOR_EXT},
    /* BLENDTEXTUREALPHA */		{E_EC, GLTF_SOURCEA, 1, GL_INTERPOLATE_EXT, 0, 0, 1, GL_TEXTURE},
    /* BLENDFACTORALPHA */		{E_EC, GLTF_SOURCEA, 1, GL_INTERPOLATE_EXT, 0, 0, 1, GL_CONSTANT_EXT},
    /* BLENDCURRENTALPHA */		{E_EC, GLTF_SOURCEA, 1, GL_INTERPOLATE_EXT, 0, 0, 1, GL_PREVIOUS_EXT},
/* supported by ARB_texture_env_dot3 or EXT_texture_env_dot3 (incompatible extensions) */
    /* DOTPRODUCT3 */			{E_A3, GLTF_SOURCEA|GLTF_SA, 1, GL_DOT3_RGB_ARB, GL_DOT3_RGBA_ARB, 0, 1}, /* ? */
    /* DOTPRODUCT3 */			{E_E3, GLTF_SOURCEA|GLTF_SA, 1, GL_DOT3_RGB_EXT, GL_DOT3_RGBA_EXT, 0, 1}  /* ? */
};

static struct {
    int c;
    GL_TEXOP *op;
} GL_tex_op[] = {
    {0,NULL}, {0,NULL},      /* DISABLE */
    {1,&GL_tex_op_data[ 0]}, /* SELECTARG1 */
    {1,&GL_tex_op_data[ 1]}, /* SELECTARG2 */
    {1,&GL_tex_op_data[ 2]}, /* MODULATE */
    {1,&GL_tex_op_data[ 5]}, /* MODULATE2X */
    {1,&GL_tex_op_data[ 6]}, /* MODULATE4X */
    {1,&GL_tex_op_data[ 3]}, /* ADD */
    {1,&GL_tex_op_data[ 7]}, /* ADDSIGNED */
    {1,&GL_tex_op_data[ 8]}, /* ADDSIGNED2X */
    {1,&GL_tex_op_data[ 4]}, /* SUBTRACT */
    {0,NULL},                /* ADDSMOOTH */
    {1,&GL_tex_op_data[ 9]}, /* BLENDDIFFUSEALPHA */
    {1,&GL_tex_op_data[10]}, /* BLENDTEXTUREALPHA */
    {1,&GL_tex_op_data[11]}, /* BLENDFACTORALPHA */
    {0,NULL},                /* BLENDTEXTUREALPHAPM */
    {1,&GL_tex_op_data[12]}, /* BLENDCURRENTALPHA */
    {0,NULL},                /* PREMODULATE */
    {0,NULL},                /* MODULATEALPHA_ADDCOLOR */
    {0,NULL},                /* MODULATECOLOR_ADDALPHA */
    {0,NULL},                /* MODULATEINVALPHA_ADDCOLOR */
    {0,NULL},                /* MODULATEINVCOLOR_ADDALPHA */
    {0,NULL},                /* BUMPENVMAP */
    {0,NULL},                /* BUMPENVMAPLUMINANCE */
    {2,&GL_tex_op_data[13]}, /* DOTPRODUCT3 */
};

static const GLint GL_D3DTA[] = {
    GL_PRIMARY_COLOR_EXT, /* diffuse */
    GL_PREVIOUS_EXT,
    GL_TEXTURE,
    GL_CONSTANT_EXT,
    GL_PRIMARY_COLOR_EXT /* specular, not supported by GL yet */
};

static const GLenum GL_mag[] = {
    0,
    GL_NEAREST,	/* POINT */
    GL_LINEAR,	/* LINEAR */
    GL_LINEAR,	/* FLATCUBIC */
    GL_LINEAR,	/* GAUSSIANCUBIC */
    GL_LINEAR	/* ANISOTROPIC */
};

static const GLenum GL_mip_min[4][4] = {
/* min: POINT                      LINEAR                    ANISOTROPIC */
    {0, 0,                         0,                        0},			/* mip: */
    {0, GL_NEAREST,                GL_LINEAR,                GL_LINEAR},		/* NONE */
    {0, GL_NEAREST_MIPMAP_NEAREST, GL_LINEAR_MIPMAP_NEAREST, GL_LINEAR_MIPMAP_NEAREST},	/* POINT */
    {0, GL_NEAREST_MIPMAP_LINEAR,  GL_LINEAR_MIPMAP_LINEAR,  GL_LINEAR_MIPMAP_LINEAR}	/* LINEAR */
};

static const GLenum GL_tex_addr[] = {
    0,
    GL_REPEAT,			/* WRAP */
    GL_MIRRORED_REPEAT_IBM,	/* MIRROR */
    GL_CLAMP_TO_EDGE,		/* CLAMP */
    GL_CLAMP_TO_BORDER_ARB	/* BORDER */
};

/* Returns whether we can use a particular extension */
static inline BOOL D3D_GL_have_tex_ext(GL_Direct3DDeviceImpl* priv, DD_GL_ext ext)
{
    if (ext == E_GL) return TRUE;

    if (HAVE_GL_EXT(priv, ext)) return TRUE;

    /* In case there are GL implementations that report ARB but not EXT */
    if (ext == EXT_texture_env_add) return HAVE_GL_EXT(priv, ARB_texture_env_add);
    if (ext == EXT_texture_env_combine) return HAVE_GL_EXT(priv, ARB_texture_env_combine);

    return FALSE;
}

/* Find the most appropriate texture operation descriptor from GL_tex_op table
 * Returns NULL if no available GL extension can support the requested operation
 */
static inline GL_TEXOP* D3D_GL_get_tex_op(GL_Direct3DDeviceImpl* priv, DWORD op)
{
    int c;
    GL_TEXOP* gop;
    c = GL_tex_op[op].c;
    gop = GL_tex_op[op].op;
    while (c) {
	if (D3D_GL_have_tex_ext(priv, gop->ext)) return gop;
	gop++;
	c--;
    }
    return NULL;
}

#define OPERAND_OFS (GL_OPERAND0_RGB_EXT - GL_SOURCE0_RGB_EXT)

/* Compute GL_OPERANDn_{RGB|ALPHA}_EXT parameter from D3D tex-op argument */
static inline GLenum GL_get_alpha_opn(DWORD dwValue) {
    if (dwValue & D3DTA_COMPLEMENT) return GL_ONE_MINUS_SRC_ALPHA;
    return GL_SRC_ALPHA;
}

static inline GLenum GL_get_color_opn(DWORD dwValue) {
    if (dwValue & D3DTA_ALPHAREPLICATE) return GL_get_alpha_opn(dwValue);
    if (dwValue & D3DTA_COMPLEMENT) return GL_ONE_MINUS_SRC_COLOR;
    return GL_SRC_COLOR;
}

/* Template for inline functions to be given to COND_TEX_ENV.
 * One such function is instantiated for every possible GL tex-op source.
 * Template takes:
 *  name = name of function
 *  flag,src = GL tex-op source (see GLTF flags above)
 *  arg1,arg2 = D3D argument fields
 *  opn_func = function to compute GL_OPERANDn_{RGB|ALPHA}_EXT parameter from D3D argument
 *  opn_def = default GL_OPERANDn_{RGB|ALPHA}_EXT if D3D arguments are not used
 * Instantiated function takes:
 *  dwStage = texture stage to take any tex-op arguments from
 *  gop = texture operation descriptor
 * Function returns:
 *  src = parameter for GL_SOURCEn_{RGB|ALPHA}_EXT
 *  opn = parameter for GL_OPERANDn_{RGB|ALPHA}_EXT
 */
#define MAKE_TEX_SRC(name,flag,src,arg1,arg2,opn_func,opn_def) \
static inline void name(IDirect3DDeviceImpl* This, DWORD dwStage, GL_TEXOP* gop, GLenum* src, GLenum* opn) \
{ \
    if (gop->dwFlags & flag) { \
	DWORD dwValue = gop->src ? \
			This->d3dp.texstagestate[dwStage][D3DTSS_##arg2] : \
			This->d3dp.texstagestate[dwStage][D3DTSS_##arg1]; \
	*src = GL_D3DTA[dwValue & D3DTA_SELECTMASK]; \
	*opn = opn_func(dwValue); \
    } else { \
	*src = gop->src; \
	*opn = opn_def; \
    } \
}

MAKE_TEX_SRC(GL_get_color_src0,GLTF_SOURCE0,src0,COLORARG1,COLORARG2,GL_get_color_opn,GL_SRC_COLOR)
MAKE_TEX_SRC(GL_get_color_src1,GLTF_SOURCE1,src1,COLORARG1,COLORARG2,GL_get_color_opn,GL_SRC_COLOR)
MAKE_TEX_SRC(GL_get_color_src2,GLTF_SOURCE2,src2,COLORARG1,COLORARG2,GL_get_color_opn,GL_SRC_ALPHA)
MAKE_TEX_SRC(GL_get_alpha_src0,GLTF_SOURCE0,src0,ALPHAARG1,ALPHAARG2,GL_get_alpha_opn,GL_SRC_ALPHA)
MAKE_TEX_SRC(GL_get_alpha_src1,GLTF_SOURCE1,src1,ALPHAARG1,ALPHAARG2,GL_get_alpha_opn,GL_SRC_ALPHA)
MAKE_TEX_SRC(GL_get_alpha_src2,GLTF_SOURCE2,src2,ALPHAARG1,ALPHAARG2,GL_get_alpha_opn,GL_SRC_ALPHA)

#undef MAKE_TEX_SRC

/* Template for inline functions to be given to COND_TEX_ARG.
 * One such function is instantiated for color, and one for alpha.
 * Template takes:
 *  name = name of function
 *  suff = GL tex-op source suffix (RGB or ALPHA)
 * Instantiated function takes:
 *  gop = texture operation descriptor
 *  no = D3D argument number (0 or 1, see GLTF flags above)
 * Function returns:
 *  which GL_SOURCEn_{RGB|ALPHA}_EXT parameter the D3D argument maps to.
 */
#define MAKE_TEX_SRC(name,suff) \
static inline GLenum name(GL_TEXOP* gop, int no) { \
    if ((gop->dwFlags & GLTF_SOURCE0) && gop->src0 == no) return GL_SOURCE0_##suff; \
    if ((gop->dwFlags & GLTF_SOURCE1) && gop->src1 == no) return GL_SOURCE1_##suff; \
    if ((gop->dwFlags & GLTF_SOURCE2) && gop->src2 == no) return GL_SOURCE2_##suff; \
    return 0; \
}

MAKE_TEX_SRC(GL_get_color_src,RGB_EXT)
MAKE_TEX_SRC(GL_get_alpha_src,ALPHA_EXT)

#undef MAKE_TEX_SRC

/* Conditional Texture Environment.
 * Sets GL tex-op parameters as required by the (newly set) texture operation.
 * Assumes:
 *  dwStage = texture stage
 * Takes:
 *  env = GL tex-op source to set
 *  gop = texture operation descriptor
 *  func = function to compute GL tex-op parameters
 */
#define COND_TEX_ENV(env, gop, func) \
    do { GLenum src, opn; \
	 func(This, dwStage, gop, &src, &opn); \
	 TRACE(" arg %04x %04x %04x\n", env, src, opn); \
	 if (src) { glTexEnvi(GL_TEXTURE_ENV, env, src); \
	  glTexEnvi(GL_TEXTURE_ENV, env + OPERAND_OFS, opn); } } while (0)

/* Conditional Texture Argument.
 * Changes GL tex-op parameters to reflect a change in D3D tex-op arguments.
 * Assumes:
 *  dwStage = texture stage
 *  dwState = D3D tex-op argument that has changed
 * Takes:
 *  base = D3D tex-op argument base (to compute D3D argument number from)
 *  gop = texture operation descriptor
 *  env_func = function to compute which GL tex-op source to change
 *  opn_func = function to compute GL_OPERANDn_{RGB|ALPHA}_EXT parameter
 */
#define COND_TEX_ARG(base, gop, env_func, opn_func) \
    do { DWORD dwValue = This->d3dp.texstagestate[dwStage][dwState]; \
	 GLenum env = env_func(gop, dwState - base), \
		src = GL_D3DTA[dwValue & D3DTA_SELECTMASK], \
		opn = opn_func(dwValue); \
	 TRACE(" arg %04x %04x %04x\n", env, src, opn); \
	 if (env) { glTexEnvi(GL_TEXTURE_ENV, env, src); \
	  glTexEnvi(GL_TEXTURE_ENV, env + OPERAND_OFS, opn); } } while (0)

static void D3D_GL_set_tex_op(IDirect3DDeviceImpl* This, DWORD dwStage, D3DTEXTURESTAGESTATETYPE dwState)
{
    GL_D3D_PRIV_VAR(priv, This);
    if (D3D_GL_have_tex_ext(priv, EXT_texture_env_combine)) {
	switch (dwState) {
	case D3DTSS_COLOROP: {
		DWORD state = This->d3dp.texstagestate[dwStage][D3DTSS_COLOROP];
		GL_TEXOP *cop = D3D_GL_get_tex_op(priv, state);
		if (state == D3DTOP_DISABLE) {
		    TRACE(" stage disable\n");
		    GL_StageDisabled(This, dwStage);
		    return;
		}
		if (!cop) {
		    TRACE(" unsupported\n");
		    return;
		}
		if (cop->dwFlags & GLTF_SA) FIXME(" special alpha\n");
		TRACE(" color mode %04x\n", cop->cmode);
		glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, cop->cmode);
		glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_EXT, cop->scale);
		COND_TEX_ENV(GL_SOURCE0_RGB_EXT, cop, GL_get_color_src0);
		COND_TEX_ENV(GL_SOURCE1_RGB_EXT, cop, GL_get_color_src1);
		COND_TEX_ENV(GL_SOURCE2_RGB_EXT, cop, GL_get_color_src2);
		GL_StageEnabled(This, dwStage);
	    }
	    break;
	case D3DTSS_ALPHAOP: {
		DWORD state = This->d3dp.texstagestate[dwStage][D3DTSS_ALPHAOP];
		GL_TEXOP *aop = D3D_GL_get_tex_op(priv, state);
		if (!aop) {
		    TRACE(" unsupported\n");
		    return;
		}
		if (aop->dwFlags & GLTF_SA) FIXME(" special alpha\n");
		TRACE(" alpha mode %04x\n", aop->cmode);
		glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA_EXT, aop->cmode);
		glTexEnvi(GL_TEXTURE_ENV, GL_ALPHA_SCALE, aop->scale);
		COND_TEX_ENV(GL_SOURCE0_ALPHA_EXT, aop, GL_get_alpha_src0);
		COND_TEX_ENV(GL_SOURCE1_ALPHA_EXT, aop, GL_get_alpha_src1);
		COND_TEX_ENV(GL_SOURCE2_ALPHA_EXT, aop, GL_get_alpha_src2);
	    }
	    break;
	case D3DTSS_COLORARG1:
	case D3DTSS_COLORARG2:
	    {
		GL_TEXOP *cop = D3D_GL_get_tex_op(priv, This->d3dp.texstagestate[dwStage][D3DTSS_COLOROP]);
		if (!cop) return;
		COND_TEX_ARG(D3DTSS_COLORARG1, cop, GL_get_color_src, GL_get_color_opn);
	    }
	    break;
	case D3DTSS_ALPHAARG1:
	case D3DTSS_ALPHAARG2:
	    {
		GL_TEXOP *aop = D3D_GL_get_tex_op(priv, This->d3dp.texstagestate[dwStage][D3DTSS_ALPHAOP]);
		if (!aop) return;
		COND_TEX_ARG(D3DTSS_ALPHAARG1, aop, GL_get_alpha_src, GL_get_alpha_opn);
	    }
	    break;
	default:
	}
	return;
    }
    if (dwState == D3DTSS_COLOROP) {
	DWORD state = This->d3dp.texstagestate[dwStage][D3DTSS_COLOROP];
	GL_TEXOP *cop = D3D_GL_get_tex_op(priv, state);
	if (state == D3DTOP_DISABLE) {
	    GL_StageDisabled(This, dwStage);
	    return;
	}
	if (!cop) {
	    TRACE(" unsupported\n");
	    return;
	}
	TRACE(" mode %04x\n", cop->cmode);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, cop->cmode);
	if (cop->dwFlags & GLTF_SA) FIXME(" special alpha\n");
	GL_StageEnabled(This, dwStage);
    }
}

#undef COND_TEX_ARG
#undef COND_TEX_ENV

static HRESULT D3D_GL_validate_tex_stage(IDirect3DDeviceImpl* This, DWORD dwStage)
{
    GL_D3D_PRIV_VAR(priv, This);
    GL_TEXOP *cop = D3D_GL_get_tex_op(priv, This->d3dp.texstagestate[dwStage][D3DTSS_COLOROP]);
    /* GL_TEXOP *aop = D3D_GL_get_tex_op(priv, This->d3dp.texstagestate[dwStage][D3DTSS_ALPHAOP]); */
    if (!cop) return D3DERR_UNSUPPORTEDCOLOROPERATION;
    if (D3D_GL_have_tex_ext(priv, EXT_texture_env_combine)) {
	/* FIXME */
	return D3D_OK;
    }
    /* FIXME */
    return D3D_OK;
}

HRESULT D3D_GL_texture_stage_state(IDirect3DDeviceImpl* This, LPD3DHAL_DP2COMMAND cmd, LPVOID data1, LPVOID data2)
{
    LPD3DHAL_DP2TEXTURESTAGESTATE dat = (LPD3DHAL_DP2TEXTURESTAGESTATE)data1;
    GL_D3D_PRIV_VAR(priv, This);
    if (dat->wStage >= priv->gl.lpCaps->wMaxTextureBlendStages) {
	WARN("unsupported texture stage\n");
	return D3D_OK;
    }
    TRACE("stage %d, state %d, value %08lx\n", dat->wStage, dat->TSState, dat->dwValue);
    GLD(priv)->glActiveTextureARB(GL_TEXTURE0_ARB + dat->wStage);
    switch (dat->TSState) {
    case D3DTSS_TEXTUREMAP:
#ifndef WIREFRAME
	if (dat->dwValue) {
#else
	if (dat->dwValue && !gWireframe) {
#endif   
            IDirectDrawSurfaceImpl *tex = OBJECT_FROM_FIELD(IDirectDrawSurfaceImpl, t, (void*)dat->dwValue);
	    GLfloat col[4];
	    if (!DIBGLDEVICE(tex)) {
		HRESULT hr = GL_PrepTexture(This, tex);
		if (FAILED(hr)) return hr;
	    }
	    glBindTexture(GL_TEXTURE_2D, TEXTURE_ID(tex));
	    if (GL_IsStageEnabled(This, dat->wStage)) glEnable(GL_TEXTURE_2D);

	    /* since I think these parameters are bound to the texture in OpenGL,
	     * but not in Direct3D, they must be reloaded for every texture change */
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			    GL_mag[This->d3dp.texstagestate[dat->wStage][D3DTSS_MAGFILTER]]);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			    GL_mip_min[This->d3dp.texstagestate[dat->wStage][D3DTSS_MIPFILTER]]
				      [This->d3dp.texstagestate[dat->wStage][D3DTSS_MINFILTER]]);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,
			    GL_tex_addr[This->d3dp.texstagestate[dat->wStage][D3DTSS_ADDRESSU]]);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,
			    GL_tex_addr[This->d3dp.texstagestate[dat->wStage][D3DTSS_ADDRESSV]]);
	    GL_RS2Col(col, This->d3dp.texstagestate[dat->wStage][D3DTSS_BORDERCOLOR]);
	    glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, col);
	    if (HAVE_GL_EXT(priv, EXT_texture_filter_anisotropic))
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT,
				This->d3dp.texstagestate[dat->wStage][D3DTSS_MAXANISOTROPY]);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL,
			    This->d3dp.texstagestate[dat->wStage][D3DTSS_MAXMIPLEVEL]);
	} else {
	    glBindTexture(GL_TEXTURE_2D, 0);
	    glDisable(GL_TEXTURE_2D);
	}
	break;
    case D3DTSS_COLOROP:
    case D3DTSS_COLORARG1:
    case D3DTSS_COLORARG2:
    case D3DTSS_ALPHAOP:
    case D3DTSS_ALPHAARG1:
    case D3DTSS_ALPHAARG2:
	D3D_GL_set_tex_op(This, dat->wStage, dat->TSState);
	break;
    case D3DTSS_TEXCOORDINDEX:
	priv->gl.glFlags |= GLF_NEW_TEXCOORD_TRANSFORM;

	if ((This->d3dp.texstagestate[dat->wStage][D3DTSS_TEXCOORDINDEX]
	     & D3DTSS_TCI_MASK) != 0)
	{
	    priv->gl.texture_flags &= ~GLTF_DEFAULT_TEXCOORD_TRANSFORM;
	}
	break;
    case D3DTSS_ADDRESS:
    case D3DTSS_ADDRESSU:
    case D3DTSS_ADDRESSV:
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,
			GL_tex_addr[This->d3dp.texstagestate[dat->wStage][D3DTSS_ADDRESSU]]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,
			GL_tex_addr[This->d3dp.texstagestate[dat->wStage][D3DTSS_ADDRESSV]]);
	break;
    case D3DTSS_BORDERCOLOR:
	{
	    GLfloat col[4];
	    GL_RS2Col(col, This->d3dp.texstagestate[dat->wStage][D3DTSS_BORDERCOLOR]);
	    glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, col);
	}
	break;
    case D3DTSS_MAGFILTER:
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			GL_mag[This->d3dp.texstagestate[dat->wStage][D3DTSS_MAGFILTER]]);
	break;
    case D3DTSS_MINFILTER:
    case D3DTSS_MIPFILTER:
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_mip_min[This->d3dp.texstagestate[dat->wStage][D3DTSS_MIPFILTER]]
				  [This->d3dp.texstagestate[dat->wStage][D3DTSS_MINFILTER]]);
	break;

    case D3DTSS_MIPMAPLODBIAS:
	/* FIXME: Reactivate this when DRI gets fixed - crasher in Mesa 3.4/3.5 gl_texture_pixels
	 * function that tries to dereference lambda[] array when lod_bias is on but mipmapping
	 * is off.
	 
	if (HAVE_GL_EXT(priv, EXT_texture_lod_bias))
	    glTexEnvf(GL_TEXTURE_FILTER_CONTROL_EXT, GL_TEXTURE_LOD_BIAS_EXT,
		      GL_RSF(dat->dwValue));
	*/ 
	break;
    case D3DTSS_MAXANISOTROPY:
	if (HAVE_GL_EXT(priv, EXT_texture_filter_anisotropic))
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT,
			    This->d3dp.texstagestate[dat->wStage][D3DTSS_MAXANISOTROPY]);
	break;
    case D3DTSS_MAXMIPLEVEL:
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, This->d3dp.texstagestate[dat->wStage][D3DTSS_MAXMIPLEVEL]);
        break; 
    case D3DTSS_TEXTURETRANSFORMFLAGS:
	priv->gl.glFlags |= GLF_NEW_TEXCOORD_TRANSFORM;

	if (This->d3dp.texstagestate[dat->wStage][D3DTSS_TEXTURETRANSFORMFLAGS]
	    != D3DTTFF_DISABLE)
	    priv->gl.texture_flags &= ~GLTF_DEFAULT_TEXCOORD_TRANSFORM;
        break; /* This is handled elsewhere - just record it for now */
    default:
	FIXME("unsupported texture stage state %d\n", dat->TSState);
	break;
    }
    CHECK_GL_ERROR("TextureStageState");
    return D3D_OK;
}

HRESULT WINAPI GL_Direct3DDevice_validate_device(LPVOID ctx, LPDWORD lpdwPasses)
{
    IDirect3DDeviceImpl* This = (IDirect3DDeviceImpl*)ctx;
    GL_D3D_PRIV_VAR(priv, This);
    unsigned u;
    HRESULT hr = D3D_OK;

    TRACE("(%p)->(%p)\n", This, lpdwPasses);
    *lpdwPasses = 1;

    for (u=0; (!hr) && u < priv->gl.lpCaps->wMaxTextureBlendStages &&
	      This->d3dp.texstagestate[u][D3DTSS_COLOROP] > D3DTOP_DISABLE; u++)
	hr = D3D_GL_validate_tex_stage(This, u);

    TRACE("returning %ld\n", hr);
    return hr;
}
